import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.Misc";

CircularArray = class();

function CircularArray:Constructor()
	self.data = { };
end

function CircularArray:realIndex(index)
	if #self.data == 0 then
		return 0;
	end
	return index % #self.data;
end

function CircularArray:get(index)
	return self.data[self:realIndex(index) + 1];
end

function CircularArray:insert(index, item)
	index = self:realIndex(index);
	table.insert(self.data, index + 1, item);
	return index;
end

function CircularArray:remove(index)
	index = self:realIndex(index);
	table.remove(self.data, index + 1);
	return index;
end

CircularQuickslot = class(Turbine.UI.Lotro.Quickslot);

function CircularQuickslot:Constructor(parent, size)
	Turbine.UI.Lotro.Quickslot.Constructor(self);

	self.parent = parent;
	self.state = 0;
	self.safe = false;

	self:SetParent(parent);
	self:SetSize(size, size);
end

function CircularQuickslot:SafeSetShortcut(shortcut)
	self.safe = true;
	local status = pcall(self.SetShortcut, self, shortcut);
	self.safe = false;
	return status;
end

function CircularQuickslot:MouseClick(args)
	self.parent:toggle();
end

function CircularQuickslot:MouseEnter(args)
	self.state = 0;
end

function CircularQuickslot:MouseLeave(args)
	self.state = 1;
end

function CircularQuickslot:MouseUp(args)
	if self.state == 1 then
		self.state = 2;
	end

	self.parent:redraw();
end

function CircularQuickslot:DragLeave(args)
	if self.state == 2 then
		self:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
	end

	self.parent:redraw();
	self.state = 0;
end

function CircularQuickslot:MouseWheel(args)
	self.parent.settings.index = self.parent.array:realIndex(self.parent.settings.index - tonumber(args.Direction));
	self.parent:redraw();
end

function CircularQuickslot:ShortcutChanged(args)
	if self.safe then
		return;
	end

	local shortcut = self:GetShortcut();
	local index;

	if shortcut == nil or shortcut:GetType() == Turbine.UI.Lotro.ShortcutType.Undefined then
		index = self.parent.array:remove(self.parent.settings.index + self.index);
		if index < self.parent.settings.index then
			self.parent.settings.index = self.parent.array:realIndex(self.parent.settings.index - 1);
		end
	else
		local item = { };
		local index;

		item.type = shortcut:GetType();
		item.data = shortcut:GetData();

		index = self.parent.settings.index + self.index;
		if self.index < 0 then
			index = index + 1;
		end
		index = self.parent.array:insert(index, item);
		if self.index ~= 0 then
			if index <= self.parent.settings.index then
				self.parent.settings.index = self.parent.array:realIndex(self.parent.settings.index + 1);
			end
		end
	end

	self.parent:redraw();
	self.parent:SaveData();
end

CircularQuickslotBar = class(Turbine.UI.Window);

function CircularQuickslotBar:Constructor(name, count, settings, data)
	Turbine.UI.Window.Constructor(self);

	self.hide = false
	self.visible = true;
	self.name = name;
	self.spacing = 0;
	self.quickslotSize = 36;
	self.quickslots = { };
	self.array = CircularArray();
	if settings ~= nil then
		self.settings = settings;
	else
		self.settings = { };
		self.settings.count = 3;
		self.settings.index = 0;
	end
	if data ~= nil then
		self.array.data = data;
	end
	if count ~= nil then
		self.settings.count = count;
	end

	self:setCount(self.settings.count);
	self:SetOpacity(1);
	self:SetWantsKeyEvents(true);
	self:SetVisible();
	self:SetZOrder(1);
end

function CircularQuickslotBar:KeyDown(args)
	if args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function CircularQuickslotBar:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end
	Turbine.UI.Window.SetVisible(self, self.visible and not self.hide);
end

function CircularQuickslotBar:IsVisible()
	return self.visible;
end

function CircularQuickslotBarLoadData(name)
	data = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_data_" .. name);
	if type(data) == "string" then
		data = loadstring("return " .. data .. ";")();
	end

	if type(data) ~= "table" then
		return nil;
	end

	local qs = Turbine.UI.Lotro.Quickslot();

	for i = #data, 1, -1 do
		if not pcall(qs.SetShortcut, qs, Turbine.UI.Lotro.Shortcut(data[i].type, data[i].data)) then
			table.remove(data, i);
		end
	end

	return data;
end

function CircularQuickslotBar:SaveData()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_data_" .. self.name, serialize(self.array.data));
end

function CircularQuickslotBarLoadSettings(name)
	local settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_settings_" .. name);
	if type(settings) == "string" then
		settings = loadstring("return " .. settings .. ";")();
	end

	if type(settings) ~= "table" then
		return nil;
	end

	if not settings.count then
		settings.count = 3;
	end

	if not settings.index then
		settings.index = 0;
	end

	return settings;
end

function CircularQuickslotBar:SaveSettings()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_settings_" .. self.name, serialize(self.settings));
end

function CircularQuickslotBar:reposition()
	local x, y = Turbine.UI.Display:GetMousePosition();
	local width = self.quickslotSize;
	local height = (self.quickslotSize + self.spacing) * #self.quickslots - self.spacing;

	self:SetPosition(x - (width / 2), y - (height / 2));
end

function CircularQuickslotBar:redraw()
	for _, v in ipairs(self.quickslots) do
		local item = self.array:get(v.index + self.settings.index);

		if item == nil then
			v:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
		else
			v:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(item.type, item.data));
		end
		v:SetVisible(true);
	end

	self:SetVisible(false);
	self:SetVisible(true);
end

function CircularQuickslotBar:setCount(newCount)
	local delta = #self.quickslots - newCount;

	if delta < 0 then
		delta = -delta;
		while delta > 0 do
			table.insert(self.quickslots, CircularQuickslot(self, self.quickslotSize));
			delta = delta - 1;
		end
	else
		local quickslot;

		while delta > 0 do
			quickslot = table.remove(self.quickslots);
			quickslot:SetVisible(false);
			quickslot:SetParent(nil);
			delta = delta - 1;
		end
	end

	for k, v in ipairs(self.quickslots) do
		v.index = math.ceil(-#self.quickslots / 2) + (k - 1);
		v:SetPosition(0, (k - 1) * (self.quickslotSize + self.spacing));
	end
	self:SetSize(self.quickslotSize, (self.quickslotSize + self.spacing) * #self.quickslots - self.spacing);

	self.settings.count = #self.quickslots;

	self:reposition();
	self:SaveSettings();
end

function CircularQuickslotBar:toggle()
	if self:IsVisible() then
		self:SetVisible(false);
	else
		self:reposition();
		self:redraw();
		self:SetVisible(true);
		self:Focus();
	end
end

function CircularQuickslotBar:random()
	self.settings.index = self.array:realIndex(math.random(#self.array.data));
end

function plugin.Unload(args)
	local names = { };

	for k, v in pairs(CircularQuickslotBars) do
		v:SaveSettings();
		v:SaveData();
		table.insert(names, k);
	end

	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_names", serialize(names));
end

CircularQuickslotCommand = class(Turbine.ShellCommand);

function CircularQuickslotCommand:Constructor(array)
	Turbine.ShellCommand.Constructor(self);

	self.array = array;
end

function CircularQuickslotCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	if args[1] ~= nil then
		if self.array[args[1]] == nil then
			local names = { };

			self.array[args[1]] = CircularQuickslotBar(args[1], tonumber(args[2]), nil, nil);

			for k in pairs(self.array) do
				table.insert(names, k);
			end
			Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_names", serialize(names));
		else
			if tonumber(args[2]) ~= nil then
				self.array[args[1]]:setCount(tonumber(args[2]));
			elseif args[2] == "random" then
				self.array[args[1]]:random();
			end
			self.array[args[1]]:toggle();
		end
	else
		Turbine.Shell.WriteLine("Usage: /cq <name> [#]");
	end
end

function CircularQuickslotCommand:GetShortHelp()
	return "Circular quickslot";
end

CircularQuickslotBars = { };

local names = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_names");
if type(names) == "string" then
	names = loadstring("return " .. names .. ";")();
end
if names ~= nil then
	for _, v in pairs(names) do
		local data = CircularQuickslotBarLoadData(v);
		local settings = CircularQuickslotBarLoadSettings(v);
		if data ~= nil and settings ~= nil then
			CircularQuickslotBars[v] = CircularQuickslotBar(v, nil, settings, data);
			CircularQuickslotBars[v]:toggle();
		end
	end
end

plugin.command = CircularQuickslotCommand(CircularQuickslotBars);
Turbine.Shell.AddCommand("circularquickslot", plugin.command);
Turbine.Shell.AddCommand("cq", plugin.command);

math.randomseed(Turbine.Engine.GetGameTime());